from pathlib import Path

import pytest

from cg.apps.demultiplex.sample_sheet.override_cycles_validator import OverrideCyclesValidator
from cg.apps.demultiplex.sample_sheet.sample_models import (
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.constants.demultiplexing import SampleSheetBcl2FastqSections, SampleSheetBCLConvertSections


@pytest.fixture
def bcl_convert_samples_similar_index1() -> list[FlowCellSampleBCLConvert]:
    """Return a list of three FlowCellSampleBCLConvert with updated indexes."""
    sample_1 = FlowCellSampleBCLConvert(
        lane=1, sample_id="ACC123", index="CAGAAGAT", index2="GCGCAAGC"
    )
    sample_2 = FlowCellSampleBCLConvert(
        lane=1, sample_id="ACC456", index="CAGAAGAG", index2="CAATGTAT"
    )
    sample_3 = FlowCellSampleBCLConvert(
        lane=2, sample_id="ACC789", index="AAGCGATA", index2="AACCGCAA"
    )
    return [sample_1, sample_2, sample_3]


@pytest.fixture
def bcl_convert_samples_similar_index2() -> list[FlowCellSampleBCLConvert]:
    """Return a list of three FlowCellSampleBCLConvert with updated indexes."""
    sample_1 = FlowCellSampleBCLConvert(
        lane=1, sample_id="ACC123", index="GCGCAAGC", index2="CAATGTAC"
    )
    sample_2 = FlowCellSampleBCLConvert(
        lane=1, sample_id="ACC456", index="CAATGTAT", index2="CAATGTAT"
    )
    sample_3 = FlowCellSampleBCLConvert(
        lane=2, sample_id="ACC789", index="AAGCGATA", index2="AACCGCAA"
    )
    return [sample_1, sample_2, sample_3]


# Sample sheet validation


@pytest.fixture
def sample_sheet_line_sample_1() -> list[str]:
    """Return the line in the sample sheet corresponding to a sample."""
    return [
        "HWHMWDMXX",
        "1",
        "ACC7628A68",
        "hg19",
        "ATTCCACACT",
        "TGGTCTTGTT",
        "814206",
        "N",
        "R1",
        "script",
        "814206",
    ]


@pytest.fixture
def sample_sheet_line_sample_2() -> list[str]:
    """Return the line in the sample sheet corresponding to a sample."""
    return [
        "HWHMWDMXX",
        "1",
        "ACC7628A1",
        "hg19",
        "AGTTAGCTGG",
        "GATGAGAATG",
        "814206",
        "N",
        "R1",
        "script",
        "814206",
    ]


@pytest.fixture
def sample_sheet_bcl2fastq_data_header() -> list[list[str]]:
    """Return the content of a Bcl2fastq sample sheet data header without samples."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        [
            SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value,
            SampleSheetBcl2FastqSections.Data.LANE.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_INTERNAL_ID_BCL2FASTQ.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_REFERENCE.value,
            SampleSheetBcl2FastqSections.Data.INDEX_1.value,
            SampleSheetBcl2FastqSections.Data.INDEX_2.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_NAME.value,
            SampleSheetBcl2FastqSections.Data.CONTROL.value,
            SampleSheetBcl2FastqSections.Data.RECIPE.value,
            SampleSheetBcl2FastqSections.Data.OPERATOR.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_PROJECT_BCL2FASTQ.value,
        ],
    ]


@pytest.fixture
def sample_sheet_bcl2fastq_data_header_with_replaced_sample_id() -> list[list[str]]:
    """Return the content of a Bcl2fastq sample sheet data header without samples."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        [
            SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value,
            SampleSheetBcl2FastqSections.Data.LANE.value,
            SampleSheetBCLConvertSections.Data.SAMPLE_INTERNAL_ID.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_REFERENCE.value,
            SampleSheetBcl2FastqSections.Data.INDEX_1.value,
            SampleSheetBcl2FastqSections.Data.INDEX_2.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_NAME.value,
            SampleSheetBcl2FastqSections.Data.CONTROL.value,
            SampleSheetBcl2FastqSections.Data.RECIPE.value,
            SampleSheetBcl2FastqSections.Data.OPERATOR.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_PROJECT_BCL2FASTQ.value,
        ],
    ]


@pytest.fixture
def sample_sheet_samples_no_header(
    sample_sheet_line_sample_1: list[str], sample_sheet_line_sample_2: list[str]
) -> list[list[str]]:
    """Return the content of a sample sheet with samples but without a sample header."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture
def valid_sample_sheet_bcl2fastq(
    sample_sheet_bcl2fastq_data_header: list[list[str]],
    sample_sheet_line_sample_1: list[str],
    sample_sheet_line_sample_2: list[str],
) -> list[list[str]]:
    """Return the content of a valid Bcl2fastq sample sheet."""
    return sample_sheet_bcl2fastq_data_header + [
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture
def sample_sheet_bcl2fastq_duplicate_same_lane(
    valid_sample_sheet_bcl2fastq: list[list[str]], sample_sheet_line_sample_2: list[str]
) -> list[list[str]]:
    """Return the content of a Bcl2fastq sample sheet with a duplicated sample in the same lane."""
    valid_sample_sheet_bcl2fastq.append(sample_sheet_line_sample_2)
    return valid_sample_sheet_bcl2fastq


@pytest.fixture
def sample_sheet_bcl2fastq_duplicate_different_lane(
    valid_sample_sheet_bcl2fastq: list[list[str]],
) -> list[list[str]]:
    """Return the content of a Bcl2fastq sample sheet with a duplicated sample in a different lane."""
    valid_sample_sheet_bcl2fastq.append(
        [
            "HWHMWDMXX",
            "2",
            "ACC7628A1",
            "hg19",
            "AGTTAGCTGG",
            "GATGAGAATG",
            "814206",
            "N",
            "R1",
            "script",
            "814206",
        ]
    )
    return valid_sample_sheet_bcl2fastq


@pytest.fixture
def novaseq6000_flow_cell_sample_1() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=1,
        SampleID="ACC7628A68",
        SampleRef="hg19",
        index="ATTCCACACT",
        index2="TGGTCTTGTT",
        SampleName="814206",
        Control="N",
        Recipe="R1",
        Operator="script",
        Project="814206",
    )


@pytest.fixture
def novaseq6000_flow_cell_sample_2() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        SampleRef="hg19",
        index="ATTCCACACT",
        index2="TGGTCTTGTT",
        SampleName="814206",
        Control="N",
        Recipe="R1",
        Operator="script",
        Project="814206",
    )


@pytest.fixture
def novaseq6000_flow_cell_sample_no_dual_index() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample without dual indexes."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        index="ATTCCACACT",
        SampleName="814206",
        Project="814206",
    )


@pytest.fixture
def novaseq6000_flow_cell_sample_before_adapt_indexes() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample without dual indexes."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        index="ATTCCACACT-TGGTCTTGTT",
        SampleName="814206",
        Project="814206",
    )


@pytest.fixture
def index1_8_nt_sequence_from_lims() -> str:
    """Return an index 1 sequence."""
    return "GTCTACAC"


@pytest.fixture
def index2_8_nt_sequence_from_lims() -> str:
    """Return an index 2 sequence."""
    return "GCCAAGGT"


@pytest.fixture
def index1_10_nt_sequence_from_lims() -> str:
    """Return an index 1 sequence."""
    return "CCGGTTCATG"


@pytest.fixture
def index2_10_nt_sequence_from_lims() -> str:
    """Return an index 2 sequence."""
    return "CAAGACGTCT"


@pytest.fixture
def raw_index_sequence(
    index1_8_nt_sequence_from_lims: str, index2_8_nt_sequence_from_lims: str
) -> str:
    """Return a raw index."""
    return f"{index1_8_nt_sequence_from_lims}-{index2_8_nt_sequence_from_lims}"


@pytest.fixture
def bcl_convert_flow_cell_sample(raw_index_sequence: str) -> FlowCellSampleBCLConvert:
    """Return a BCL Convert sample."""
    return FlowCellSampleBCLConvert(lane=1, index=raw_index_sequence, sample_id="ACC123")


@pytest.fixture
def bcl_convert_sample_sheet_path(illumina_demultiplexed_runs_directory):
    return Path(
        illumina_demultiplexed_runs_directory,
        "230504_A00689_0804_BHY7FFDRX2",
        "SampleSheet.csv",
    )


@pytest.fixture
def override_cycles_validator() -> OverrideCyclesValidator:
    """Return an override cycles validator without any initialised attribute."""
    return OverrideCyclesValidator()


@pytest.fixture
def processed_flow_cell_sample_8_index(
    index1_8_nt_sequence_from_lims: str, index2_8_nt_sequence_from_lims: str
) -> dict[str, str]:
    """Return a BCL Convert sample with processed 8-nt indexes and no override cycles."""
    return {
        "Lane": 1,
        "Sample_ID": "ACC123",
        "Index": index1_8_nt_sequence_from_lims,
        "Index2": index2_8_nt_sequence_from_lims,
    }


@pytest.fixture
def processed_flow_cell_sample_10_index(
    index1_10_nt_sequence_from_lims: str, index2_10_nt_sequence_from_lims: str
) -> dict[str, str]:
    """Return a BCL Convert sample with processed 10-nt indexes and no override cycles."""
    return {
        "Lane": 1,
        "Sample_ID": "ACC123",
        "Index": index1_10_nt_sequence_from_lims,
        "Index2": index2_10_nt_sequence_from_lims,
    }


@pytest.fixture
def forward_index2_cycle_processed_flow_cell_8_nt_sample(
    processed_flow_cell_sample_8_index: dict[str, str],
) -> dict[str, str]:
    """Return a BCL Convert sample with processed 8-nt indexes and forward index 2 cycle."""
    processed_flow_cell_sample_8_index["OverrideCycles"] = "Y151;I8N2;I8N2;Y151"
    return processed_flow_cell_sample_8_index


@pytest.fixture
def reverse_index2_cycle_processed_flow_cell_8_nt_sample(
    processed_flow_cell_sample_8_index: dict[str, str],
) -> dict[str, str]:
    """Return a BCL Convert sample with processed 8-nt indexes and reversed index 2 cycle."""
    processed_flow_cell_sample_8_index["OverrideCycles"] = "Y151;I8N2;N2I8;Y151"
    return processed_flow_cell_sample_8_index


@pytest.fixture
def processed_flow_cell_10_nt_sample(
    processed_flow_cell_sample_10_index: dict[str, str],
) -> dict[str, str]:
    """Return a BCL Convert sample with processed 10-nt indexes."""
    processed_flow_cell_sample_10_index["OverrideCycles"] = "Y151;I10;I10;Y151"
    return processed_flow_cell_sample_10_index



from cg.constants import DEFAULT_CAPTURE_KIT, Workflow
from cg.constants.constants import AnalysisType
from cg.constants.gene_panel import GENOME_BUILD_37
from cg.constants.pedigree import Pedigree
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.mip.mip_analysis import MipAnalysis
from cg.store.models import CaseSample
from cg.utils import Process


class MipDNAAnalysisAPI(MipAnalysisAPI):
    def __init__(self, config: CGConfig, workflow: Workflow = Workflow.MIP_DNA):
        super().__init__(config, workflow)

    @property
    def root(self) -> str:
        return self.config.mip_rd_dna.root

    @property
    def conda_binary(self) -> str:
        return self.config.mip_rd_dna.conda_binary

    @property
    def conda_env(self) -> str:
        return self.config.mip_rd_dna.conda_env

    @property
    def mip_workflow(self) -> str:
        return self.config.mip_rd_dna.workflow

    @property
    def script(self) -> str:
        return self.config.mip_rd_dna.script

    @property
    def use_read_count_threshold(self) -> bool:
        return True

    @property
    def process(self) -> Process:
        if not self._process:
            self._process = Process(
                binary=f"{self.script} {self.mip_workflow}",
                conda_binary=f"{self.conda_binary}" if self.conda_binary else None,
                config=self.config.mip_rd_dna.mip_config,
                environment=self.conda_env,
            )
        return self._process

    def config_sample(
        self, link_obj: CaseSample, panel_bed: str | None
    ) -> dict[str, str | int | None]:
        """Return config sample data."""
        sample_data: dict[str, str | int] = self.get_sample_data(link_obj=link_obj)
        if sample_data["analysis_type"] == AnalysisType.WHOLE_GENOME_SEQUENCING:
            sample_data["capture_kit"]: str = panel_bed or DEFAULT_CAPTURE_KIT
        else:
            sample_data["capture_kit"]: str | None = panel_bed or self.get_target_bed_from_lims(
                case_id=link_obj.case.internal_id
            )
        if link_obj.mother:
            sample_data[Pedigree.MOTHER.value]: str = link_obj.mother.internal_id
        if link_obj.father:
            sample_data[Pedigree.FATHER.value]: str = link_obj.father.internal_id
        return sample_data

    def get_gene_panel(self, case_id: str) -> list[str]:
        """Create and return the aggregated gene panel file."""
        return self._get_gene_panel(case_id=case_id, genome_build=GENOME_BUILD_37)

    def get_managed_variants(self) -> list[str]:
        """Create and return the managed variants."""
        return self._get_managed_variants(genome_build=GENOME_BUILD_37)

    def get_genome_build(self, analysis_metadata: MipAnalysis) -> str:
        """Return the reference genome build version of a MIP-DNA analysis."""
        return analysis_metadata.genome_build



"""Set case attributes in the status database."""

import logging

import click

from cg.constants import DataDelivery, Priority, Workflow
from cg.constants.constants import CaseActions
from cg.models.cg_config import CGConfig
from cg.store.models import Case, Customer, Panel
from cg.store.store import Store
from cg.utils.click.EnumChoice import EnumChoice

LOG = logging.getLogger(__name__)


@click.command("case")
@click.option("-a", "--action", type=click.Choice(CaseActions.actions()), help="update case action")
@click.option("-c", "--customer-id", type=click.STRING, help="update customer")
@click.option(
    "-d",
    "--data-analysis",
    "data_analysis",
    type=EnumChoice(Workflow),
    help="Update case data analysis",
)
@click.option(
    "-dd",
    "--data-delivery",
    "data_delivery",
    type=EnumChoice(DataDelivery),
    help="Update case data delivery",
)
@click.option("-g", "--panel", "panel_abbreviations", multiple=True, help="update gene panels")
@click.option(
    "-p", "--priority", type=EnumChoice(Priority, use_value=False), help="update priority"
)
@click.argument("case_id")
@click.pass_obj
def set_case(
    context: CGConfig,
    action: str | None,
    data_analysis: Workflow | None,
    data_delivery: DataDelivery | None,
    priority: Priority | None,
    panel_abbreviations: tuple[str] | None,
    case_id: str,
    customer_id: str | None,
):
    """Update information about a case."""

    options: list[str] = [
        action,
        panel_abbreviations,
        priority,
        customer_id,
        data_analysis,
        data_delivery,
    ]
    abort_on_empty_options(options=options)

    status_db: Store = context.status_db
    case: Case = get_case(case_id=case_id, status_db=status_db)

    if action:
        update_action(case=case, action=action)

    if customer_id:
        update_customer(case=case, customer_id=customer_id, status_db=status_db)

    if data_analysis:
        update_data_analysis(case=case, data_analysis=data_analysis)

    if data_delivery:
        update_data_delivery(case=case, data_delivery=data_delivery)

    if panel_abbreviations:
        update_panels(case=case, panel_abbreviations=panel_abbreviations, status_db=status_db)

    if priority:
        update_priority(case=case, priority=priority)

    status_db.session.commit()


def abort_on_empty_options(options: list[str]) -> None:
    if not any(options):
        LOG.error("Nothing to change")
        raise click.Abort


def get_case(case_id: str, status_db: Store) -> Case:
    case: Case = status_db.get_case_by_internal_id(internal_id=case_id)

    if case is None:
        LOG.error(f"Can't find case {case_id}")
        raise click.Abort

    return case


def update_action(case: Case, action: str) -> None:
    """Update case action."""
    LOG.info(f"Update action: {case.action or 'NA'} -> {action}")
    case.action = action


def update_customer(case: Case, customer_id: str, status_db: Store) -> None:
    customer_obj: Customer = status_db.get_customer_by_internal_id(customer_internal_id=customer_id)

    if customer_obj is None:
        LOG.error(f"Unknown customer: {customer_id}")
        raise click.Abort

    LOG.info(f"Update customer: {case.customer.internal_id} -> {customer_id}")
    case.customer = customer_obj


def update_data_analysis(case: Case, data_analysis: Workflow) -> None:
    LOG.info(f"Update data_analysis: {case.data_analysis or 'NA'} -> {data_analysis}")
    case.data_analysis = data_analysis


def update_data_delivery(case: Case, data_delivery: DataDelivery) -> None:
    LOG.info(f"Update data_delivery: {case.data_delivery or 'NA'} -> {data_delivery}")
    case.data_delivery = data_delivery


def update_panels(case: Case, panel_abbreviations: list[str], status_db: Store) -> None:
    for panel_abbreviation in panel_abbreviations:
        panel: Panel = status_db.get_panel_by_abbreviation(abbreviation=panel_abbreviation)
        if panel is None:
            LOG.error(f"unknown gene panel: {panel_abbreviation}")
            raise click.Abort
    LOG.info(f"Update panels: {', '.join(case.panels)} -> {', '.join(panel_abbreviations)}")
    case.panels = panel_abbreviations


def update_priority(case: Case, priority: Priority) -> None:
    """Update case priority."""
    LOG.info(f"Update priority: {case.priority.name} -> {priority.name}")
    case.priority = priority
